/*
 * LIBOIL - Library of Optimized Inner Loops
 * Copyright (c) 2003,2004 David A. Schleef <ds@schleef.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <liboil/liboilfunction.h>
#include <liboil/liboildebug.h>
#include <liboil/liboilcpu.h>
#include <liboil/liboilfault.h>
#include <liboil/liboilutils.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <setjmp.h>
#include <signal.h>
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#include <time.h>

#if defined(__FreeBSD__) || defined(__APPLE__)
#include <sys/types.h>
#include <sys/sysctl.h>
#endif

#ifdef __sun
#include <sys/auxv.h>
#endif

/***** i386, amd64 *****/

#if defined(__sun)
#define USE_I386_GETISAX
#else
#define USE_I386_CPUID
#endif


#ifdef USE_I386_CPUINFO
static void
oil_cpu_i386_getflags_cpuinfo (char *cpuinfo)
{
  char *cpuinfo_flags;
  char **flags;
  char **f;

  cpuinfo_flags = get_tag_value (cpuinfo, "flags");
  if (cpuinfo_flags == NULL) {
    free (cpuinfo);
    return;
  }

  flags = strsplit(cpuinfo_flags);
  for (f = flags; *f; f++) {
    if (strcmp (*f, "cmov") == 0) {
      OIL_DEBUG ("cpu flag %s", *f);
      oil_cpu_flags |= OIL_IMPL_FLAG_CMOV;
    }
    if (strcmp (*f, "mmx") == 0) {
      OIL_DEBUG ("cpu flag %s", *f);
      oil_cpu_flags |= OIL_IMPL_FLAG_MMX;
    }
    if (strcmp (*f, "sse") == 0) {
      OIL_DEBUG ("cpu flag %s", *f);
      oil_cpu_flags |= OIL_IMPL_FLAG_SSE;
    }
    if (strcmp (*f, "mmxext") == 0) {
      OIL_DEBUG ("cpu flag %s", *f);
      oil_cpu_flags |= OIL_IMPL_FLAG_MMXEXT;
    }
    if (strcmp (*f, "sse2") == 0) {
      OIL_DEBUG ("cpu flag %s", *f);
      oil_cpu_flags |= OIL_IMPL_FLAG_SSE2;
      oil_cpu_flags |= OIL_IMPL_FLAG_MMXEXT;
    }
    if (strcmp (*f, "3dnow") == 0) {
      OIL_DEBUG ("cpu flag %s", *f);
      oil_cpu_flags |= OIL_IMPL_FLAG_3DNOW;
    }
    if (strcmp (*f, "3dnowext") == 0) {
      OIL_DEBUG ("cpu flag %s", *f);
      oil_cpu_flags |= OIL_IMPL_FLAG_3DNOWEXT;
    }
    if (strcmp (*f, "sse3") == 0) {
      OIL_DEBUG ("cpu flag %s", *f);
      oil_cpu_flags |= OIL_IMPL_FLAG_SSE3;
      oil_cpu_flags |= OIL_IMPL_FLAG_SSE2;
      oil_cpu_flags |= OIL_IMPL_FLAG_MMXEXT;
    }
    if (strcmp (*f, "ssse3") == 0) {
      OIL_DEBUG ("cpu flag %s", *f);
      oil_cpu_flags |= OIL_IMPL_FLAG_SSSE3;
      oil_cpu_flags |= OIL_IMPL_FLAG_SSE3;
      oil_cpu_flags |= OIL_IMPL_FLAG_SSE2;
      oil_cpu_flags |= OIL_IMPL_FLAG_MMXEXT;
    }

    free (*f);
  }
  free (flags);
  free (cpuinfo);
  free (cpuinfo_flags);
}
#endif

#ifdef HAVE_GCC_ASM
static unsigned long
oil_profile_stamp_rdtsc(void)
{
	unsigned long ts;
	__asm__ __volatile__("rdtsc\n" : "=a" (ts) : : "edx");
	return ts;
}
#endif

#ifdef USE_I386_CPUID
#ifdef _MSC_VER // get CPU INFORMATION

#define X_MMX_Technology    0x00800000  // OIL_IMPL_FLAG_MMX
#define X_SSE_Extensions    0x02000000  // OIL_IMPL_FLAG_SSE
#define X_SSE2_Extensions   0x04000000  // OIL_IMPL_FLAG_SSE2

typedef struct tagALL_CPU_INFO {
    OilImplFlag oilflag;
    char CPUString[0x20];
    char CPUBrandString[0x40];
    int nSteppingID;
    int nModel;
    int nFamily;
    int nProcessorType;
    int nExtendedmodel;
    int nExtendedfamily;
    int nBrandIndex;
    int nCLFLUSHcachelinesize;
    int nLogicalProcessors;
    int nAPICPhysicalID;
    int nFeatureInfo;
    int nCacheLineSize;
    int nL2Associativity;
    int nCacheSizeK;
    int nPhysicalAddress;
    int nVirtualAddress;

    int    bSSE3Instructions;
    int    bMONITOR_MWAIT;
    int    bCPLQualifiedDebugStore;
    int    bVirtualMachineExtensions;
    int    bEnhancedIntelSpeedStepTechnology;
    int    bThermalMonitor2;
    int    bSupplementalSSE3;
    int    bL1ContextID;
    int    bCMPXCHG16B;
    int    bxTPRUpdateControl;
    int    bPerfDebugCapabilityMSR;
    int    bSSE41Extensions;
    int    bSSE42Extensions;
    int    bPOPCNT;

    int    bMultithreading;

    int    bLAHF_SAHFAvailable;
    int    bCmpLegacy;
    int    bSVM;
    int    bExtApicSpace;
    int    bAltMovCr8;
    int    bLZCNT;
    int    bSSE4A;
    int    bMisalignedSSE;
    int    bPREFETCH;
    int    bSKINITandDEV;
    int    bSYSCALL_SYSRETAvailable;
    int    bExecuteDisableBitAvailable;
    int    bMMXExtensions;
    int    bFFXSR;
    int    b1GBSupport;
    int    bRDTSCP;
    int    b64Available;
    int    b3DNowExt;
    int    b3DNow;
    int    bNestedPaging;
    int    bLBRVisualization;
    int    bFP128;
    int    bMOVOptimization;
    unsigned    nIds, nExIds;
    int    CPUInfo[4];
    int    extCPUInfo[4];
    int    iTotalInfos;
    int *  pCPUInfo;
}ALL_CPU_INFO, * PALL_CPU_INFO;

/* ===========================================================================
    PALL_CPU_INFO get_all_CPU_information(void)
    No information is output - just collected,
    and moved into simple BOOL values where necessary/desireable
   =========================================================================== */
PALL_CPU_INFO get_all_CPU_information(void)
{
    unsigned  i, off, nExId;
    int * pCPUI;
    PALL_CPU_INFO pcpi = malloc(sizeof(ALL_CPU_INFO));

    if ( !pcpi )
        return NULL;

    memset(pcpi,0,sizeof(ALL_CPU_INFO));

    // __cpuid with an InfoType argument of 0 returns the number of
    // valid Ids in CPUInfo[0] and the CPU identification string in
    // the other three array elements. The CPU identification string is
    // not in linear order. The code below arranges the information 
    // in a human readable form.
    __cpuid(&pcpi->CPUInfo[0], 0);
    pcpi->nIds                   = pcpi->CPUInfo[0];
    *((int*)pcpi->CPUString)     = pcpi->CPUInfo[1];
    *((int*)(pcpi->CPUString+4)) = pcpi->CPUInfo[3];
    *((int*)(pcpi->CPUString+8)) = pcpi->CPUInfo[2];

    // Calling __cpuid with 0x80000000 as the InfoType argument
    // gets the number of valid extended IDs.
    __cpuid(&pcpi->extCPUInfo[0], 0x80000000);
    pcpi->nExIds = pcpi->extCPUInfo[0];
    nExId = 0;
    for ( i = 0x80000000; i <= pcpi->nExIds; ++i )
        nExId++;
    pcpi->iTotalInfos = sizeof(int) * 4 * (pcpi->nIds + nExId + 4);
    pcpi->pCPUInfo = malloc(pcpi->iTotalInfos);
    if (!pcpi->pCPUInfo) {
        free(pcpi);
        return NULL;
    }
    // Get the information associated with each valid Id
    off = 0;
    for ( i = 0; i <= pcpi->nIds; ++i )
    {
        pCPUI = &pcpi->pCPUInfo[off];
        __cpuid(pCPUI, i);

        // Interpret CPU feature information.
        if  (i == 1)
        {
            pcpi->nSteppingID                       = pCPUI[0] & 0xf;
            pcpi->nModel                            = (pCPUI[0] >> 4) & 0xf;
            pcpi->nFamily                           = (pCPUI[0] >> 8) & 0xf;
            pcpi->nProcessorType                    = (pCPUI[0] >> 12) & 0x3;
            pcpi->nExtendedmodel                    = (pCPUI[0] >> 16) & 0xf;
            pcpi->nExtendedfamily                   = (pCPUI[0] >> 20) & 0xff;
            pcpi->nBrandIndex                       = pCPUI[1] & 0xff;
            pcpi->nCLFLUSHcachelinesize             = ((pCPUI[1] >> 8) & 0xff) * 8;
            pcpi->nLogicalProcessors                = ((pCPUI[1] >> 16) & 0xff);
            pcpi->nAPICPhysicalID                   = (pCPUI[1] >> 24) & 0xff;
            pcpi->bSSE3Instructions                 = (pCPUI[2] & 0x1) ? 1 : 0;
            pcpi->bMONITOR_MWAIT                    = (pCPUI[2] & 0x8) ? 1 : 0;
            pcpi->bCPLQualifiedDebugStore           = (pCPUI[2] & 0x10) ? 1 : 0;
            pcpi->bVirtualMachineExtensions         = (pCPUI[2] & 0x20) ? 1 : 0;
            pcpi->bEnhancedIntelSpeedStepTechnology = (pCPUI[2] & 0x80) ? 1 : 0;
            pcpi->bThermalMonitor2                  = (pCPUI[2] & 0x100) ? 1 : 0;
            pcpi->bSupplementalSSE3                 = (pCPUI[2] & 0x200) ? 1 : 0;
            pcpi->bL1ContextID                      = (pCPUI[2] & 0x300) ? 1 : 0;
            pcpi->bCMPXCHG16B                       = (pCPUI[2] & 0x2000) ? 1 : 0;
            pcpi->bxTPRUpdateControl                = (pCPUI[2] & 0x4000) ? 1 : 0;
            pcpi->bPerfDebugCapabilityMSR           = (pCPUI[2] & 0x8000) ? 1 : 0;
            pcpi->bSSE41Extensions                  = (pCPUI[2] & 0x80000) ? 1 : 0;
            pcpi->bSSE42Extensions                  = (pCPUI[2] & 0x100000) ? 1 : 0;
            pcpi->bPOPCNT                           = (pCPUI[2] & 0x800000) ? 1 : 0;
            pcpi->nFeatureInfo                      = pCPUI[3];
            pcpi->bMultithreading                   = (pcpi->nFeatureInfo & (1 << 28)) ? 1 : 0;
        }
        off += 4;
    }

    // Get the information associated with each extended ID.
    for ( i = 0x80000000; i <= pcpi->nExIds; ++i )
    {
        pCPUI = &pcpi->pCPUInfo[off];
        __cpuid(pCPUI, i);
        if  (i == 0x80000001)
        {
            pcpi->bLAHF_SAHFAvailable         = (pCPUI[2] & 0x1) ? 1 : 0;
            pcpi->bCmpLegacy                  = (pCPUI[2] & 0x2) ? 1 : 0;
            pcpi->bSVM                        = (pCPUI[2] & 0x4) ? 1 : 0;
            pcpi->bExtApicSpace               = (pCPUI[2] & 0x8) ? 1 : 0;
            pcpi->bAltMovCr8                  = (pCPUI[2] & 0x10) ? 1 : 0;
            pcpi->bLZCNT                      = (pCPUI[2] & 0x20) ? 1 : 0;
            pcpi->bSSE4A                      = (pCPUI[2] & 0x40) ? 1 : 0;
            pcpi->bMisalignedSSE              = (pCPUI[2] & 0x80) ? 1 : 0;
            pcpi->bPREFETCH                   = (pCPUI[2] & 0x100) ? 1 : 0;
            pcpi->bSKINITandDEV               = (pCPUI[2] & 0x1000) ? 1 : 0;
            pcpi->bSYSCALL_SYSRETAvailable    = (pCPUI[3] & 0x800) ? 1 : 0;
            pcpi->bExecuteDisableBitAvailable = (pCPUI[3] & 0x10000) ? 1 : 0;
            pcpi->bMMXExtensions              = (pCPUI[3] & 0x40000) ? 1 : 0;
            pcpi->bFFXSR                      = (pCPUI[3] & 0x200000) ? 1 : 0;
            pcpi->b1GBSupport                 = (pCPUI[3] & 0x400000) ? 1 : 0;
            pcpi->bRDTSCP                     = (pCPUI[3] & 0x8000000) ? 1 : 0;
            pcpi->b64Available                = (pCPUI[3] & 0x20000000) ? 1 : 0;
            pcpi->b3DNowExt                   = (pCPUI[3] & 0x40000000) ? 1 : 0;
            pcpi->b3DNow                      = (pCPUI[3] & 0x80000000) ? 1 : 0;
        }

        // Interpret CPU brand string and cache information.
        if  (i == 0x80000002)
            memcpy(pcpi->CPUBrandString, pCPUI, sizeof(pcpi->CPUInfo));
        else if  (i == 0x80000003)
            memcpy(pcpi->CPUBrandString + 16, pCPUI, sizeof(pcpi->CPUInfo));
        else if  (i == 0x80000004)
            memcpy(pcpi->CPUBrandString + 32, pCPUI, sizeof(pcpi->CPUInfo));
        else if  (i == 0x80000006)
        {
            pcpi->nCacheLineSize   = pCPUI[2] & 0xff;
            pcpi->nL2Associativity = (pCPUI[2] >> 12) & 0xf;
            pcpi->nCacheSizeK      = (pCPUI[2] >> 16) & 0xffff;
        }
        else if  (i == 0x80000008)
        {
           pcpi->nPhysicalAddress = pCPUI[0] & 0xff;
           pcpi->nVirtualAddress  = (pCPUI[0] >> 8) & 0xff;
        }
        else if  (i == 0x8000000A)
        {
            pcpi->bNestedPaging     = (pCPUI[3] & 0x1) ? 1 : 0;
            pcpi->bLBRVisualization = (pCPUI[3] & 0x2) ? 1 : 0;
        }
        else if  (i == 0x8000001A)
        {
            pcpi->bFP128           = (pCPUI[0] & 0x1) ? 1 : 0;
            pcpi->bMOVOptimization = (pCPUI[0] & 0x2) ? 1 : 0;
        }
        off += 4;
    }
    return pcpi;
}

void kill_all_CPU_information(PALL_CPU_INFO pcpi)
{
    if( pcpi ) {
        if (pcpi->pCPUInfo)
            free(pcpi->pCPUInfo);
        free(pcpi);
    }
}

void set_oil_flag( PALL_CPU_INFO pcpi )
{
    if (pcpi) {
        //unsigned int bit;
        // OIL_IMPL_FLAG_REF ???
        // OIL_IMPL_FLAG_OPT ???
        // OIL_IMPL_FLAG_ASM ???
        // OIL_IMPL_FLAG_DISABLED ???
        // OIL_IMPL_FLAG_CMOV ???

        // OIL_IMPL_FLAG_MMX = "MMX Technology"
        //bit = Get_Feature_Bit( OIL_IMPL_FLAG_MMX, "MMX Technology" );
        //if ( pcpi->nFeatureInfo & bit)
        if ( pcpi->nFeatureInfo & X_MMX_Technology)
            pcpi->oilflag |= OIL_IMPL_FLAG_MMX;

        // OIL_IMPL_FLAG_SSE = "SSE Extensions"
        //bit = Get_Feature_Bit( OIL_IMPL_FLAG_SSE, "SSE Extensions" );
        //if ( pcpi->nFeatureInfo & bit)
        if ( pcpi->nFeatureInfo & X_SSE_Extensions)
            pcpi->oilflag |= OIL_IMPL_FLAG_SSE;

        // OIL_IMPL_FLAG_MMXEXT
        if (pcpi->bMMXExtensions)
            pcpi->oilflag |= OIL_IMPL_FLAG_MMXEXT;

        // OIL_IMPL_FLAG_SSE2 = "SSE2 Extensions"
        //bit = Get_Feature_Bit( OIL_IMPL_FLAG_SSE2, "SSE2 Extensions" );
        if ( pcpi->nFeatureInfo & X_SSE2_Extensions)
            pcpi->oilflag |= OIL_IMPL_FLAG_SSE2;

        // OIL_IMPL_FLAG_3DNOW
        if ( pcpi->b3DNow )
            pcpi->oilflag |= OIL_IMPL_FLAG_3DNOW;

        // OIL_IMPL_FLAG_3DNOWEXT
        if ( pcpi->b3DNowExt )
            pcpi->oilflag |= OIL_IMPL_FLAG_3DNOWEXT;

        // OIL_IMPL_FLAG_SSE3
        if  (pcpi->bSSE3Instructions)
            pcpi->oilflag |= OIL_IMPL_FLAG_SSE3;

        // OIL_IMPL_FLAG_ALTIVEC ???
        // OIL_IMPL_FLAG_EDSP ???
        // OIL_IMPL_FLAG_ARM6 ???
        // OIL_IMPL_FLAG_VFP ???

        // OIL_IMPL_FLAG_SSSE3
        if  (pcpi->bSupplementalSSE3)
            pcpi->oilflag |= OIL_IMPL_FLAG_SSSE3;

    }
}

static void
oil_cpu_detect_cpuid(void)
{
    PALL_CPU_INFO pcpi = get_all_CPU_information();
    if ( pcpi ) {
        set_oil_flag( pcpi );
        oil_cpu_flags |= pcpi->oilflag;
        kill_all_CPU_information(pcpi);
    }
}
#else // !_MSC_VER
#ifdef __i386__
static void
get_cpuid (uint32_t op, uint32_t *a, uint32_t *b, uint32_t *c, uint32_t *d)
{
  __asm__ (
      "  pushl %%ebx\n"
      "  cpuid\n"
      "  mov %%ebx, %%esi\n"
      "  popl %%ebx\n"
      : "=a" (*a), "=S" (*b), "=c" (*c), "=d" (*d)
      : "0" (op));
}
#endif

#ifdef __amd64__
static void
get_cpuid (uint32_t op, uint32_t *a, uint32_t *b, uint32_t *c, uint32_t *d)
{
  __asm__ (
      "  pushq %%rbx\n"
      "  cpuid\n"
      "  mov %%ebx, %%esi\n"
      "  popq %%rbx\n"
      : "=a" (*a), "=S" (*b), "=c" (*c), "=d" (*d)
      : "0" (op));
}
#endif

static void
test_cpuid (void *ignored)
{
  uint32_t eax, ebx, ecx, edx;

  get_cpuid (0x00000000, &eax, &ebx, &ecx, &edx);
}

static void
oil_cpu_detect_cpuid (void)
{
  uint32_t eax, ebx, ecx, edx;
  uint32_t level;
  char vendor[13] = { 0 };
  int ret;

  oil_fault_check_enable ();
  ret = oil_fault_check_try(test_cpuid, NULL);
  oil_fault_check_disable ();
  if (!ret) {
    /* CPU thinks cpuid is an illegal instruction. */
    return;
  }

  get_cpuid (0x00000000, &level, (uint32_t *)(vendor+0),
      (uint32_t *)(vendor+8), (uint32_t *)(vendor+4));

  OIL_DEBUG("cpuid %d %s", level, vendor);

  if (level < 1) {
    return;
  }

  get_cpuid (0x00000001, &eax, &ebx, &ecx, &edx);

#ifdef HAVE_GCC_ASM
  if (edx & (1<<4)) {
    _oil_profile_stamp = oil_profile_stamp_rdtsc;
  }
#endif

  /* Intel flags */
  if (edx & (1<<15)) {
    oil_cpu_flags |= OIL_IMPL_FLAG_CMOV;
  }
  if (edx & (1<<23)) {
    oil_cpu_flags |= OIL_IMPL_FLAG_MMX;
  }
  if (edx & (1<<25)) {
    oil_cpu_flags |= OIL_IMPL_FLAG_SSE;
  }
  if (edx & (1<<26)) {
    oil_cpu_flags |= OIL_IMPL_FLAG_SSE2;
    oil_cpu_flags |= OIL_IMPL_FLAG_MMXEXT;
  }
  if (ecx & (1<<0)) {
    oil_cpu_flags |= OIL_IMPL_FLAG_SSE3;
  }
  
  if (memcmp (vendor, "AuthenticAMD", 12) == 0) {
    get_cpuid (0x80000001, &eax, &ebx, &ecx, &edx);

    /* AMD flags */
    if (edx & (1<<22)) {
      oil_cpu_flags |= OIL_IMPL_FLAG_MMXEXT;
    }
    if (edx & (1<<31)) {
      oil_cpu_flags |= OIL_IMPL_FLAG_3DNOW;
    }
    if (edx & (1<<30)) {
      oil_cpu_flags |= OIL_IMPL_FLAG_3DNOWEXT;
    }

    get_cpuid (0x80000005, &eax, &ebx, &ecx, &edx);

    OIL_INFO("L1 D-cache: %d kbytes, %d-way, %d lines/tag, %d line size",
        (ecx>>24)&0xff, (ecx>>16)&0xff, (ecx>>8)&0xff, ecx&0xff);
    OIL_INFO("L1 I-cache: %d kbytes, %d-way, %d lines/tag, %d line size",
        (edx>>24)&0xff, (edx>>16)&0xff, (edx>>8)&0xff, edx&0xff);

    get_cpuid (0x80000006, &eax, &ebx, &ecx, &edx);
    OIL_INFO("L2 cache: %d kbytes, %d assoc, %d lines/tag, %d line size",
        (ecx>>16)&0xffff, (ecx>>12)&0xf, (ecx>>8)&0xf, ecx&0xff);
  }
}
#endif // _MSC_VER y/n
#endif

#ifdef USE_I386_GETISAX
static void
oil_cpu_detect_getisax (void)
{
  uint_t ui;

  getisax (&ui, 1);

  if (ui & AV_386_CMOV) {
     oil_cpu_flags |= OIL_IMPL_FLAG_CMOV;
  }
  if (ui & AV_386_MMX) {
     oil_cpu_flags |= OIL_IMPL_FLAG_MMX;
  }
  if (ui & AV_386_SSE) {
     oil_cpu_flags |= OIL_IMPL_FLAG_SSE;
  }
  if (ui & AV_386_SSE2) {
     oil_cpu_flags |= OIL_IMPL_FLAG_SSE2;
     oil_cpu_flags |= OIL_IMPL_FLAG_MMXEXT;
  }
  if (ui & AV_386_SSE3) {
     oil_cpu_flags |= OIL_IMPL_FLAG_SSE3;
  }
  if (ui & AV_386_AMD_3DNow) {
    oil_cpu_flags |= OIL_IMPL_FLAG_3DNOW;
  }
  if (ui & AV_386_AMD_3DNowx) {
    oil_cpu_flags |= OIL_IMPL_FLAG_3DNOWEXT;
  }
  if (ui & AV_386_AMD_MMX) {
    oil_cpu_flags |= OIL_IMPL_FLAG_MMXEXT;
  }
}
#endif

/* Reduce the set of CPU capabilities detected by whatever detection mechanism
 * was chosen, according to kernel limitations.  SSE requires kernel support for
 * use.
 */
static void
oil_cpu_detect_kernel_support (void)
{
#if defined(__APPLE__)
  int ret, enabled;
  size_t len;

  len = sizeof(enabled);
  ret = sysctlbyname("hw.optional.sse", &enabled, &len, NULL, 0);
  if (ret || !enabled) {
    oil_cpu_flags &= ~(OIL_IMPL_FLAG_SSE | OIL_IMPL_FLAG_SSE2 |
		       OIL_IMPL_FLAG_MMXEXT | OIL_IMPL_FLAG_SSE3);
  }
#elif defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
  int ret, enabled;
  size_t len;

  len = sizeof(enabled);
  ret = sysctlbyname("hw.instruction_sse", &enabled, &len, NULL, 0);
  if (ret || !enabled) {
    oil_cpu_flags &= ~(OIL_IMPL_FLAG_SSE | OIL_IMPL_FLAG_SSE2 |
		       OIL_IMPL_FLAG_MMXEXT | OIL_IMPL_FLAG_SSE3);
  }
#elif defined(__linux__)
  /*
   * Might also want to grow a check for the old RedHat + Linux 2.2
   * unmasked SSE FPU exception bug.  Other than that, if /proc/cpuinfo
   * reported SSE, then it's safe.
   */
#elif defined(__sun) || defined(__NetBSD__) || defined(__OpenBSD__)
  /* Solaris/NetBSD/OpenBSD are OK */
#else
   
  OIL_WARNING("Operating system is not known to support SSE.  "
      "Assuming it does, which might cause problems");
#if 0
  oil_cpu_flags &= ~(OIL_IMPL_FLAG_SSE | OIL_IMPL_FLAG_SSE2 |
		     OIL_IMPL_FLAG_MMXEXT | OIL_IMPL_FLAG_SSE3);
#endif
#endif
}

void
oil_cpu_detect_arch(void)
{
#ifdef USE_I386_CPUID
  oil_cpu_detect_cpuid ();
#endif
#ifdef USE_I386_GETISAX
  oil_cpu_detect_getisax ();
#endif
#ifdef USE_I386_CPUINFO
  oil_cpu_detect_cpuinfo ();
#endif

  oil_cpu_detect_kernel_support ();
}


